@@ -31,6 +31,7 @@
 #include <string.h>
 #include <sys/ioctl.h>
 #include <arpa/inet.h>
+#include <sys/time.h>
 
 #include <netpacket/packet.h>
 #include <netinet/if_ether.h>
@@ -72,6 +73,7 @@ typedef enum _dhcp_client_state {
 	INFORMATION_REQ,
 	SOLICITATION,
 	REQUEST,
+	CONFIRM,
 	RENEW,
 	REBIND,
 	RELEASE,
@@ -130,6 +132,8 @@ struct _GDHCPClient {
 	gpointer rebind_data;
 	GDHCPClientEventFunc release_cb;
 	gpointer release_data;
+	GDHCPClientEventFunc confirm_cb;
+	gpointer confirm_data;
 	char *last_address;
 	unsigned char *duid;
 	int duid_len;
@@ -143,6 +147,8 @@ struct _GDHCPClient {
 	time_t last_renew;
 	time_t last_rebind;
 	time_t expire;
+	gboolean retransmit;
+	struct timeval start_time;
 };
 
 static inline void debug(GDHCPClient *client, const char *format, ...)
@@ -229,7 +235,7 @@ static void add_dhcpv6_send_options(GDHCPClient *dhcp_client,
 		.ptr_buf = ptr_buf
 	};
 
-	if (dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client->type != G_DHCP_IPV6)
 		return;
 
 	g_hash_table_foreach(dhcp_client->send_value_hash,
@@ -249,16 +255,30 @@ static void copy_option(uint8_t *buf, uint16_t code, uint16_t len,
 		memcpy(&buf[4], msg, len);
 }
 
+static int32_t get_time_diff(struct timeval *tv)
+{
+	struct timeval now;
+	int32_t hsec;
+
+	gettimeofday(&now, NULL);
+
+	hsec = (now.tv_sec - tv->tv_sec) * 100;
+	hsec += (now.tv_usec - tv->tv_usec) / 10000;
+
+	return hsec;
+}
+
 static void add_dhcpv6_request_options(GDHCPClient *dhcp_client,
 				struct dhcpv6_packet *packet,
 				unsigned char *buf, int max_buf,
 				unsigned char **ptr_buf)
 {
 	GList *list;
-	uint16_t code;
+	uint16_t code, value;
+	int32_t diff;
 	int len;
 
-	if (dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client->type != G_DHCP_IPV6)
 		return;
 
 	for (list = dhcp_client->request_list; list; list = list->next) {
@@ -313,6 +333,31 @@ static void add_dhcpv6_request_options(GDHCPClient *dhcp_client,
 		case G_DHCPV6_ORO:
 			break;
 
+		case G_DHCPV6_ELAPSED_TIME:
+			if (dhcp_client->retransmit == FALSE) {
+				/*
+				 * Initial message, elapsed time is 0.
+				 */
+				diff = 0;
+			} else {
+				diff = get_time_diff(&dhcp_client->start_time);
+				if (diff < 0 || diff > 0xffff)
+					diff = 0xffff;
+			}
+
+			len = 2 + 2 + 2;
+			if ((*ptr_buf + len) > (buf + max_buf)) {
+				debug(dhcp_client, "Too long dhcpv6 message "
+					"when writing elapsed time option");
+				return;
+			}
+
+			value = htons((uint16_t)diff);
+			copy_option(*ptr_buf, G_DHCPV6_ELAPSED_TIME,
+				2, (uint8_t *)&value);
+			(*ptr_buf) += len;
+			break;
+
 		case G_DHCPV6_DNS_SERVERS:
 			break;
 
@@ -363,11 +408,11 @@ static int send_discover(GDHCPClient *dhcp_client, uint32_t requested)
 	packet.secs = dhcp_attempt_secs(dhcp_client);
 
 	if (requested)
-		dhcp_add_simple_option(&packet, DHCP_REQUESTED_IP, requested);
+		dhcp_add_option_uint32(&packet, DHCP_REQUESTED_IP, requested);
 
 	/* Explicitly saying that we want RFC-compliant packets helps
 	 * some buggy DHCP servers to NOT send bigger packets */
-	dhcp_add_simple_option(&packet, DHCP_MAX_SIZE, htons(576));
+	dhcp_add_option_uint16(&packet, DHCP_MAX_SIZE, 576);
 
 	add_request_options(dhcp_client, &packet);
 
@@ -389,9 +434,10 @@ static int send_select(GDHCPClient *dhcp_client)
 	packet.xid = dhcp_client->xid;
 	packet.secs = dhcp_attempt_secs(dhcp_client);
 
-	dhcp_add_simple_option(&packet, DHCP_REQUESTED_IP,
-					dhcp_client->requested_ip);
-	dhcp_add_simple_option(&packet, DHCP_SERVER_ID, dhcp_client->server_ip);
+	dhcp_add_option_uint32(&packet, DHCP_REQUESTED_IP,
+						dhcp_client->requested_ip);
+	dhcp_add_option_uint32(&packet, DHCP_SERVER_ID,
+						dhcp_client->server_ip);
 
 	add_request_options(dhcp_client, &packet);
 
@@ -410,7 +456,7 @@ static int send_renew(GDHCPClient *dhcp_client)
 
 	init_packet(dhcp_client , &packet, DHCPREQUEST);
 	packet.xid = dhcp_client->xid;
-	packet.ciaddr = dhcp_client->requested_ip;
+	packet.ciaddr = htonl(dhcp_client->requested_ip);
 
 	add_request_options(dhcp_client, &packet);
 
@@ -429,7 +475,7 @@ static int send_rebound(GDHCPClient *dhcp_client)
 
 	init_packet(dhcp_client , &packet, DHCPREQUEST);
 	packet.xid = dhcp_client->xid;
-	packet.ciaddr = dhcp_client->requested_ip;
+	packet.ciaddr = htonl(dhcp_client->requested_ip);
 
 	add_request_options(dhcp_client, &packet);
 
@@ -449,9 +495,9 @@ static int send_release(GDHCPClient *dhcp_client,
 
 	init_packet(dhcp_client, &packet, DHCPRELEASE);
 	packet.xid = rand();
-	packet.ciaddr = ciaddr;
+	packet.ciaddr = htonl(ciaddr);
 
-	dhcp_add_simple_option(&packet, DHCP_SERVER_ID, server);
+	dhcp_add_option_uint32(&packet, DHCP_SERVER_ID, server);
 
 	return dhcp_send_kernel_packet(&packet, ciaddr, CLIENT_PORT,
 						server, SERVER_PORT);
@@ -567,6 +613,22 @@ done:
 	close(sk);
 }
 
+void g_dhcpv6_client_set_retransmit(GDHCPClient *dhcp_client)
+{
+	if (dhcp_client == NULL)
+		return;
+
+	dhcp_client->retransmit = TRUE;
+}
+
+void g_dhcpv6_client_clear_retransmit(GDHCPClient *dhcp_client)
+{
+	if (dhcp_client == NULL)
+		return;
+
+	dhcp_client->retransmit = FALSE;
+}
+
 int g_dhcpv6_create_duid(GDHCPDuidType duid_type, int index, int type,
 			unsigned char **duid, int *duid_len)
 {
@@ -612,7 +674,7 @@ int g_dhcpv6_create_duid(GDHCPDuidType duid_type, int index, int type,
 int g_dhcpv6_client_set_duid(GDHCPClient *dhcp_client, unsigned char *duid,
 			int duid_len)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return -EINVAL;
 
 	g_free(dhcp_client->duid);
@@ -625,7 +687,7 @@ int g_dhcpv6_client_set_duid(GDHCPClient *dhcp_client, unsigned char *duid,
 
 uint32_t g_dhcpv6_client_get_iaid(GDHCPClient *dhcp_client)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return 0;
 
 	return dhcp_client->iaid;
@@ -648,7 +710,7 @@ int g_dhcpv6_client_get_timeouts(GDHCPClient *dhcp_client,
 				time_t *last_renew, time_t *last_rebind,
 				time_t *expire)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return -EINVAL;
 
 	if (T1 != NULL)
@@ -700,7 +762,7 @@ static void put_iaid(GDHCPClient *dhcp_client, int index, uint8_t *buf)
 
 int g_dhcpv6_client_set_ia(GDHCPClient *dhcp_client, int index,
 			int code, uint32_t *T1, uint32_t *T2,
-			gboolean add_iaaddr)
+			gboolean add_iaaddr, const char *ia_na)
 {
 	if (code == G_DHCPV6_IA_TA) {
 		uint8_t ia_options[4];
@@ -712,13 +774,29 @@ int g_dhcpv6_client_set_ia(GDHCPClient *dhcp_client, int index,
 					ia_options, sizeof(ia_options));
 
 	} else if (code == G_DHCPV6_IA_NA) {
+		struct in6_addr addr;
 
 		g_dhcp_client_set_request(dhcp_client, G_DHCPV6_IA_NA);
 
-		if (add_iaaddr == TRUE) {
+		/*
+		 * If caller has specified the IPv6 address it wishes to
+		 * to use (ia_na != NULL and address is valid), then send
+		 * the address to server.
+		 * If caller did not specify the address (ia_na == NULL) and
+		 * if the current address is not set, then we should not send
+		 * the address sub-option.
+		 */
+		if (add_iaaddr == TRUE && ((ia_na == NULL &&
+			IN6_IS_ADDR_UNSPECIFIED(&dhcp_client->ia_na) == FALSE)
+			|| (ia_na != NULL &&
+				inet_pton(AF_INET6, ia_na, &addr) == 1))) {
 #define IAADDR_LEN (16+4+4)
 			uint8_t ia_options[4+4+4+2+2+IAADDR_LEN];
 
+			if (ia_na != NULL)
+				memcpy(&dhcp_client->ia_na, &addr,
+						sizeof(struct in6_addr));
+
 			put_iaid(dhcp_client, index, ia_options);
 
 			if (T1 != NULL) {
@@ -780,6 +858,8 @@ int g_dhcpv6_client_set_oro(GDHCPClient *dhcp_client, int args, ...)
 
 	g_dhcpv6_client_set_send(dhcp_client, G_DHCPV6_ORO, values, len);
 
+	g_free(values);
+
 	return 0;
 }
 
@@ -794,13 +874,23 @@ static int send_dhcpv6_msg(GDHCPClient *dhcp_client, int type, char *msg)
 	packet = (struct dhcpv6_packet *)&buf[0];
 	ptr = buf + sizeof(struct dhcpv6_packet);
 
-	debug(dhcp_client, "sending DHCPv6 %s message", msg);
-
 	init_packet(dhcp_client, packet, type);
 
-	dhcp_client->xid = packet->transaction_id[0] << 16 |
-			packet->transaction_id[1] << 8 |
-			packet->transaction_id[2];
+	if (dhcp_client->retransmit == FALSE) {
+		dhcp_client->xid = packet->transaction_id[0] << 16 |
+				packet->transaction_id[1] << 8 |
+				packet->transaction_id[2];
+		gettimeofday(&dhcp_client->start_time, NULL);
+	} else {
+		packet->transaction_id[0] = dhcp_client->xid >> 16;
+		packet->transaction_id[1] = dhcp_client->xid >> 8 ;
+		packet->transaction_id[2] = dhcp_client->xid;
+	}
+
+	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_ELAPSED_TIME);
+
+	debug(dhcp_client, "sending DHCPv6 %s message xid 0x%04x", msg,
+							dhcp_client->xid);
 
 	max_buf = MAX_DHCPV6_PKT_SIZE - sizeof(struct dhcpv6_packet);
 
@@ -824,6 +914,11 @@ static int send_dhcpv6_request(GDHCPClient *dhcp_client)
 	return send_dhcpv6_msg(dhcp_client, DHCPV6_REQUEST, "request");
 }
 
+static int send_dhcpv6_confirm(GDHCPClient *dhcp_client)
+{
+	return send_dhcpv6_msg(dhcp_client, DHCPV6_CONFIRM, "confirm");
+}
+
 static int send_dhcpv6_renew(GDHCPClient *dhcp_client)
 {
 	return send_dhcpv6_msg(dhcp_client, DHCPV6_RENEW, "renew");
@@ -1134,7 +1229,7 @@ static int ipv4ll_recv_arp_packet(GDHCPClient *dhcp_client)
 			arp.arp_op != htons(ARPOP_REQUEST))
 		return -EINVAL;
 
-	ip_requested = ntohl(dhcp_client->requested_ip);
+	ip_requested = htonl(dhcp_client->requested_ip);
 	source_conflict = !memcmp(arp.arp_spa, &ip_requested,
 						sizeof(ip_requested));
 
@@ -1333,15 +1428,14 @@ static void start_request(GDHCPClient *dhcp_client)
 
 static uint32_t get_lease(struct dhcp_packet *packet)
 {
-	uint8_t *option_u8;
+	uint8_t *option;
 	uint32_t lease_seconds;
 
-	option_u8 = dhcp_get_option(packet, DHCP_LEASE_TIME);
-	if (option_u8 == NULL)
+	option = dhcp_get_option(packet, DHCP_LEASE_TIME);
+	if (option == NULL)
 		return 3600;
 
-	lease_seconds = dhcp_get_unaligned((uint32_t *) option_u8);
-	lease_seconds = ntohl(lease_seconds);
+	lease_seconds = get_be32(option);
 	/* paranoia: must not be prone to overflows */
 	lease_seconds &= 0x0fffffff;
 	if (lease_seconds < 10)
@@ -1361,6 +1455,7 @@ static void restart_dhcp(GDHCPClient *dhcp_client, int retry_times)
 
 	dhcp_client->retry_times = retry_times;
 	dhcp_client->requested_ip = 0;
+	dhcp_client->state = INIT_SELECTING;
 	switch_listening_mode(dhcp_client, L2);
 
 	g_dhcp_client_start(dhcp_client, dhcp_client->last_address);
@@ -1507,7 +1602,7 @@ static char *malloc_option_value_string(uint8_t *option, GDHCPOptionType type)
 		return NULL;
 	upper_length = len_of_option_as_string[type] *
 			((unsigned)len / (unsigned)optlen);
-	dest = ret = malloc(upper_length + 1);
+	dest = ret = g_malloc(upper_length + 1);
 	if (ret == NULL)
 		return NULL;
 
@@ -1517,16 +1612,13 @@ static char *malloc_option_value_string(uint8_t *option, GDHCPOptionType type)
 			dest += sprint_nip(dest, "", option);
 			break;
 		case OPTION_U16: {
-			uint16_t val_u16 = dhcp_get_unaligned(
-						(uint16_t *) option);
-			dest += sprintf(dest, "%u", ntohs(val_u16));
+			uint16_t val_u16 = get_be16(option);
+			dest += sprintf(dest, "%u", val_u16);
 			break;
 		}
 		case OPTION_U32: {
-			uint32_t val_u32 = dhcp_get_unaligned(
-						(uint32_t *) option);
-			dest += sprintf(dest, type == OPTION_U32 ? "%lu" :
-					"%ld", (unsigned long) ntohl(val_u32));
+			uint32_t val_u32 = get_be32(option);
+			dest += sprintf(dest, "%u", val_u32);
 			break;
 		}
 		case OPTION_STRING:
@@ -1665,7 +1757,7 @@ static GList *get_addresses(GDHCPClient *dhcp_client,
 
 	if (addr_count > 0 && st == 0) {
 		/* We only support one address atm */
-		char str[INET6_ADDRSTRLEN + 1];
+		char addr_str[INET6_ADDRSTRLEN + 1];
 
 		if (preferred > valid)
 			/* RFC 3315, 22.6 */
@@ -1674,11 +1766,11 @@ static GList *get_addresses(GDHCPClient *dhcp_client,
 		dhcp_client->T1 = T1;
 		dhcp_client->T2 = T2;
 
-		inet_ntop(AF_INET6, &addr, str, INET6_ADDRSTRLEN);
+		inet_ntop(AF_INET6, &addr, addr_str, INET6_ADDRSTRLEN);
 		debug(dhcp_client, "count %d addr %s T1 %u T2 %u",
-			addr_count, str, T1, T2);
+			addr_count, addr_str, T1, T2);
 
-		list = g_list_append(list, g_strdup(str));
+		list = g_list_append(list, g_strdup(addr_str));
 
 		if (code == G_DHCPV6_IA_NA)
 			memcpy(&dhcp_client->ia_na, &addr,
@@ -1820,9 +1912,10 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 	GDHCPClient *dhcp_client = user_data;
 	struct dhcp_packet packet;
 	struct dhcpv6_packet *packet6 = NULL;
-	uint8_t *message_type = NULL, *client_id = NULL, *option_u8,
+	uint8_t *message_type = NULL, *client_id = NULL, *option,
 		*server_id = NULL;
 	uint16_t option_len = 0, status = 0;
+	uint32_t xid = 0;
 	gpointer pkt;
 	unsigned char buf[MAX_DHCPV6_PKT_SIZE];
 	uint16_t pkt_len = 0;
@@ -1839,6 +1932,8 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 
 	pkt = &packet;
 
+	dhcp_client->status_code = 0;
+
 	if (dhcp_client->listen_mode == L2)
 		re = dhcp_recv_l2_packet(&packet,
 					dhcp_client->listener_sockfd);
@@ -1848,9 +1943,14 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 						dhcp_client->listener_sockfd);
 			pkt_len = re;
 			pkt = packet6;
-		} else
+			xid = packet6->transaction_id[0] << 16 |
+				packet6->transaction_id[1] << 8 |
+				packet6->transaction_id[2];
+		} else {
 			re = dhcp_recv_l3_packet(&packet,
 						dhcp_client->listener_sockfd);
+			xid = packet.xid;
+		}
 	} else if (dhcp_client->listen_mode == L_ARP) {
 		ipv4ll_recv_arp_packet(dhcp_client);
 		return TRUE;
@@ -1881,15 +1981,15 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 			return TRUE;
 		}
 
-		option_u8 = dhcpv6_get_option(packet6, pkt_len,
+		option = dhcpv6_get_option(packet6, pkt_len,
 				G_DHCPV6_STATUS_CODE, &option_len, NULL);
-		if (option_u8 != 0 && option_len > 0) {
-			status = option_u8[0]<<8 | option_u8[1];
+		if (option != 0 && option_len > 0) {
+			status = option[0]<<8 | option[1];
 			if (status != 0) {
 				debug(dhcp_client, "error code %d", status);
 				if (option_len > 2) {
 					gchar *txt = g_strndup(
-						(gchar *)&option_u8[2],
+						(gchar *)&option[2],
 						option_len - 2);
 					debug(dhcp_client, "error text: %s",
 						txt);
@@ -1897,9 +1997,7 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 				}
 			}
 			dhcp_client->status_code = status;
-		} else
-			dhcp_client->status_code = 0;
-
+		}
 	} else {
 		message_type = dhcp_get_option(&packet, DHCP_MESSAGE_TYPE);
 		if (message_type == NULL)
@@ -1910,8 +2008,8 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 		/* No message type / client id option, ignore package */
 		return TRUE;
 
-	debug(dhcp_client, "received DHCP packet (current state %d)",
-							dhcp_client->state);
+	debug(dhcp_client, "received DHCP packet xid 0x%04x "
+			"(current state %d)", xid, dhcp_client->state);
 
 	switch (dhcp_client->state) {
 	case INIT_SELECTING:
@@ -1922,10 +2020,9 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 		dhcp_client->timeout = 0;
 		dhcp_client->retry_times = 0;
 
-		option_u8 = dhcp_get_option(&packet, DHCP_SERVER_ID);
-		dhcp_client->server_ip =
-				dhcp_get_unaligned((uint32_t *) option_u8);
-		dhcp_client->requested_ip = packet.yiaddr;
+		option = dhcp_get_option(&packet, DHCP_SERVER_ID);
+		dhcp_client->server_ip = get_be32(option);
+		dhcp_client->requested_ip = ntohl(packet.yiaddr);
 
 		dhcp_client->state = REQUESTING;
 
@@ -2036,6 +2133,7 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 	case RENEW:
 	case REBIND:
 	case RELEASE:
+	case CONFIRM:
 		if (dhcp_client->type != G_DHCP_IPV6)
 			return TRUE;
 
@@ -2059,7 +2157,6 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 
 		switch_listening_mode(dhcp_client, L_NONE);
 
-		dhcp_client->status_code = 0;
 		get_dhcpv6_request(dhcp_client, packet6, pkt_len,
 						&dhcp_client->status_code);
 
@@ -2092,6 +2189,30 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 					dhcp_client->release_data);
 			return TRUE;
 		}
+		if (dhcp_client->confirm_cb != NULL) {
+			count = 0;
+			server_id = dhcpv6_get_option(packet6, pkt_len,
+						G_DHCPV6_SERVERID, &option_len,
+						&count);
+			if (server_id == NULL || count != 1 ||
+							option_len == 0) {
+				/* RFC 3315, 15.10 */
+				debug(dhcp_client,
+					"confirm server duid error, "
+					"discarding msg %p/%d/%d",
+					server_id, option_len, count);
+				return TRUE;
+			}
+			dhcp_client->server_duid = g_try_malloc(option_len);
+			if (dhcp_client->server_duid == NULL)
+				return TRUE;
+			memcpy(dhcp_client->server_duid, server_id, option_len);
+			dhcp_client->server_duid_len = option_len;
+
+			dhcp_client->confirm_cb(dhcp_client,
+						dhcp_client->confirm_data);
+			return TRUE;
+		}
 		break;
 	default:
 		break;
@@ -2216,6 +2337,16 @@ int g_dhcp_client_start(GDHCPClient *dhcp_client, const char *last_address)
 			}
 			send_dhcpv6_request(dhcp_client);
 
+		} else if (dhcp_client->confirm_cb) {
+			dhcp_client->state = CONFIRM;
+			re = switch_listening_mode(dhcp_client, L3);
+			if (re != 0) {
+				switch_listening_mode(dhcp_client, L_NONE);
+				dhcp_client->state = 0;
+				return re;
+			}
+			send_dhcpv6_confirm(dhcp_client);
+
 		} else if (dhcp_client->renew_cb) {
 			dhcp_client->state = RENEW;
 			re = switch_listening_mode(dhcp_client, L3);
@@ -2361,47 +2492,53 @@ void g_dhcp_client_register_event(GDHCPClient *dhcp_client,
 		dhcp_client->address_conflict_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_INFORMATION_REQ:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->information_req_cb = func;
 		dhcp_client->information_req_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_SOLICITATION:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->solicitation_cb = func;
 		dhcp_client->solicitation_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_ADVERTISE:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->advertise_cb = func;
 		dhcp_client->advertise_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_REQUEST:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->request_cb = func;
 		dhcp_client->request_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_RENEW:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->renew_cb = func;
 		dhcp_client->renew_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_REBIND:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->rebind_cb = func;
 		dhcp_client->rebind_data = data;
 		return;
 	case G_DHCP_CLIENT_EVENT_RELEASE:
-		if (dhcp_client->type == G_DHCP_IPV4)
+		if (dhcp_client->type != G_DHCP_IPV6)
 			return;
 		dhcp_client->release_cb = func;
 		dhcp_client->release_data = data;
 		return;
+	case G_DHCP_CLIENT_EVENT_CONFIRM:
+		if (dhcp_client->type != G_DHCP_IPV6)
+			return;
+		dhcp_client->confirm_cb = func;
+		dhcp_client->confirm_data = data;
+		return;
 	}
 }
 
@@ -2440,6 +2577,7 @@ char *g_dhcp_client_get_netmask(GDHCPClient *dhcp_client)
 	case INFORMATION_REQ:
 	case SOLICITATION:
 	case REQUEST:
+	case CONFIRM:
 	case RENEW:
 	case REBIND:
 	case RELEASE:
@@ -2577,7 +2715,7 @@ void g_dhcpv6_client_set_send(GDHCPClient *dhcp_client,
 
 void g_dhcpv6_client_reset_renew(GDHCPClient *dhcp_client)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return;
 
 	dhcp_client->last_renew = time(NULL);
@@ -2585,7 +2723,7 @@ void g_dhcpv6_client_reset_renew(GDHCPClient *dhcp_client)
 
 void g_dhcpv6_client_reset_rebind(GDHCPClient *dhcp_client)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return;
 
 	dhcp_client->last_rebind = time(NULL);
@@ -2593,7 +2731,7 @@ void g_dhcpv6_client_reset_rebind(GDHCPClient *dhcp_client)
 
 void g_dhcpv6_client_set_expire(GDHCPClient *dhcp_client, uint32_t timeout)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return;
 
 	dhcp_client->expire = time(NULL) + timeout;
@@ -2601,7 +2739,7 @@ void g_dhcpv6_client_set_expire(GDHCPClient *dhcp_client, uint32_t timeout)
 
 uint16_t g_dhcpv6_client_get_status(GDHCPClient *dhcp_client)
 {
-	if (dhcp_client == NULL || dhcp_client->type == G_DHCP_IPV4)
+	if (dhcp_client == NULL || dhcp_client->type != G_DHCP_IPV6)
 		return 0;
 
 	return dhcp_client->status_code;
