@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -39,12 +39,11 @@ enum connman_service_type {
 	CONNMAN_SERVICE_TYPE_SYSTEM    = 1,
 	CONNMAN_SERVICE_TYPE_ETHERNET  = 2,
 	CONNMAN_SERVICE_TYPE_WIFI      = 3,
-	CONNMAN_SERVICE_TYPE_WIMAX     = 4,
-	CONNMAN_SERVICE_TYPE_BLUETOOTH = 5,
-	CONNMAN_SERVICE_TYPE_CELLULAR  = 6,
-	CONNMAN_SERVICE_TYPE_GPS       = 7,
-	CONNMAN_SERVICE_TYPE_VPN       = 8,
-	CONNMAN_SERVICE_TYPE_GADGET    = 9,
+	CONNMAN_SERVICE_TYPE_BLUETOOTH = 4,
+	CONNMAN_SERVICE_TYPE_CELLULAR  = 5,
+	CONNMAN_SERVICE_TYPE_GPS       = 6,
+	CONNMAN_SERVICE_TYPE_VPN       = 7,
+	CONNMAN_SERVICE_TYPE_GADGET    = 8,
 };
 
 enum connman_service_security {
@@ -87,6 +86,7 @@ enum connman_service_proxy_method {
 };
 
 struct connman_service;
+struct connman_network;
 
 struct connman_service *connman_service_create(void);
 
@@ -107,6 +107,7 @@ char *connman_service_get_interface(struct connman_service *service);
 
 const char *connman_service_get_domainname(struct connman_service *service);
 char **connman_service_get_nameservers(struct connman_service *service);
+char **connman_service_get_timeservers_config(struct connman_service *service);
 char **connman_service_get_timeservers(struct connman_service *service);
 void connman_service_set_proxy_method(struct connman_service *service, enum connman_service_proxy_method method);
 enum connman_service_proxy_method connman_service_get_proxy_method(struct connman_service *service);
@@ -114,6 +115,14 @@ char **connman_service_get_proxy_servers(struct connman_service *service);
 char **connman_service_get_proxy_excludes(struct connman_service *service);
 const char *connman_service_get_proxy_url(struct connman_service *service);
 const char *connman_service_get_proxy_autoconfig(struct connman_service *service);
+connman_bool_t connman_service_get_favorite(struct connman_service *service);
+
+struct connman_service *connman_service_lookup_from_network(struct connman_network *network);
+
+void connman_service_create_ip4config(struct connman_service *service,
+								int index);
+void connman_service_create_ip6config(struct connman_service *service,
+								int index);
 
 #ifdef __cplusplus
 }
