@@ -76,19 +76,16 @@ struct connman_config {
 	char *ident;
 	char *name;
 	char *description;
-	connman_bool_t protected;
 	GHashTable *service_table;
 };
 
 static GHashTable *config_table = NULL;
-static GSList *protected_services = NULL;
 
 static connman_bool_t cleanup = FALSE;
 
 /* Definition of possible strings in the .config files */
 #define CONFIG_KEY_NAME                "Name"
 #define CONFIG_KEY_DESC                "Description"
-#define CONFIG_KEY_PROT                "Protected"
 
 #define SERVICE_KEY_TYPE               "Type"
 #define SERVICE_KEY_NAME               "Name"
@@ -116,7 +113,6 @@ static connman_bool_t cleanup = FALSE;
 static const char *config_possible_keys[] = {
 	CONFIG_KEY_NAME,
 	CONFIG_KEY_DESC,
-	CONFIG_KEY_PROT,
 	NULL,
 };
 
@@ -172,9 +168,6 @@ static void unregister_service(gpointer data)
 	connman_info("Removing service configuration %s",
 						config_service->ident);
 
-	protected_services = g_slist_remove(protected_services,
-						config_service);
-
 	if (config_service->virtual == TRUE)
 		goto free_only;
 
@@ -274,34 +267,6 @@ static void check_keys(GKeyFile *keyfile, const char *group,
 	g_strfreev(avail_keys);
 }
 
-static connman_bool_t
-is_protected_service(struct connman_config_service *service)
-{
-	GSList *list;
-
-	DBG("ident %s", service->ident);
-
-	for (list = protected_services; list; list = list->next) {
-		struct connman_config_service *s = list->data;
-
-		if (g_strcmp0(s->type, service->type) != 0)
-			continue;
-
-		if (s->ssid == NULL || service->ssid == NULL)
-			continue;
-
-		if (s->ssid_len != service->ssid_len)
-			continue;
-
-		if (g_strcmp0(service->type, "wifi") == 0 &&
-			strncmp(s->ssid, service->ssid, s->ssid_len) == 0) {
-			return TRUE;
-		}
-	}
-
-	return FALSE;
-}
-
 static int check_family(const char *address, int expected_family)
 {
 	int family;
@@ -392,6 +357,18 @@ out:
 	return err;
 }
 
+static connman_bool_t check_address(char *address_str, char **address)
+{
+	if (g_ascii_strcasecmp(address_str, "auto") == 0 ||
+			g_ascii_strcasecmp(address_str, "dhcp") == 0 ||
+			g_ascii_strcasecmp(address_str, "off") == 0) {
+		*address = address_str;
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 static connman_bool_t load_service_generic(GKeyFile *keyfile,
 			const char *group, struct connman_config *config,
 			struct connman_config_service *service)
@@ -401,7 +378,7 @@ static connman_bool_t load_service_generic(GKeyFile *keyfile,
 	gsize length;
 
 	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_IPv4, NULL);
-	if (str != NULL) {
+	if (str != NULL && check_address(str, &service->ipv4_address) == TRUE) {
 		mask = NULL;
 
 		if (parse_address(str, AF_INET, &service->ipv4_address,
@@ -436,7 +413,7 @@ static connman_bool_t load_service_generic(GKeyFile *keyfile,
 	}
 
 	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_IPv6, NULL);
-	if (str != NULL) {
+	if (str != NULL && check_address(str, &service->ipv6_address) == TRUE) {
 		long int value;
 		char *ptr;
 
@@ -628,11 +605,6 @@ static connman_bool_t load_service(GKeyFile *keyfile, const char *group,
 		service->ssid_len = ssid_len;
 	}
 
-	if (is_protected_service(service) == TRUE) {
-		connman_error("Trying to provision a protected service");
-		goto err;
-	}
-
 	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_EAP, NULL);
 	if (str != NULL) {
 		g_free(service->eap);
@@ -700,10 +672,6 @@ static connman_bool_t load_service(GKeyFile *keyfile, const char *group,
 		g_hash_table_insert(config->service_table, service->ident,
 					service);
 
-	if (config->protected == TRUE)
-		protected_services =
-			g_slist_prepend(protected_services, service);
-
 	connman_info("Adding service configuration %s", service->ident);
 
 	return TRUE;
@@ -743,8 +711,6 @@ static connman_bool_t load_service_from_keyfile(GKeyFile *keyfile,
 
 static int load_config(struct connman_config *config)
 {
-	GError *error = NULL;
-	gboolean protected;
 	GKeyFile *keyfile;
 	char *str;
 
@@ -769,14 +735,6 @@ static int load_config(struct connman_config *config)
 		config->description = str;
 	}
 
-	protected = g_key_file_get_boolean(keyfile, "global",
-					CONFIG_KEY_PROT, &error);
-	if (error == NULL)
-		config->protected = protected;
-	else
-		config->protected = TRUE;
-	g_clear_error(&error);
-
 	if (load_service_from_keyfile(keyfile, config) == FALSE)
 		connman_warn("Config file %s/%s.config does not contain any "
 			"configuration that can be provisioned!",
@@ -1110,7 +1068,17 @@ static void provision_service(gpointer key, gpointer value,
 			return;
 	}
 
-	if (config->ipv6_address != NULL) {
+	if (config->ipv6_address == NULL) {
+		connman_network_set_ipv6_method(network,
+						CONNMAN_IPCONFIG_METHOD_AUTO);
+	} else if (g_ascii_strcasecmp(config->ipv6_address, "off") == 0) {
+		connman_network_set_ipv6_method(network,
+						CONNMAN_IPCONFIG_METHOD_OFF);
+	} else if (g_ascii_strcasecmp(config->ipv6_address, "auto") == 0 ||
+			g_ascii_strcasecmp(config->ipv6_address, "dhcp") == 0) {
+		connman_network_set_ipv6_method(network,
+						CONNMAN_IPCONFIG_METHOD_AUTO);
+	} else {
 		struct connman_ipaddress *address;
 
 		if (config->ipv6_prefix_length == 0 ||
@@ -1146,7 +1114,17 @@ static void provision_service(gpointer key, gpointer value,
 							config->ipv6_privacy);
 	}
 
-	if (config->ipv4_address != NULL) {
+	if (config->ipv4_address == NULL) {
+		connman_network_set_ipv4_method(network,
+						CONNMAN_IPCONFIG_METHOD_DHCP);
+	} else if (g_ascii_strcasecmp(config->ipv4_address, "off") == 0) {
+		connman_network_set_ipv4_method(network,
+						CONNMAN_IPCONFIG_METHOD_OFF);
+	} else if (g_ascii_strcasecmp(config->ipv4_address, "auto") == 0 ||
+			g_ascii_strcasecmp(config->ipv4_address, "dhcp") == 0) {
+		connman_network_set_ipv4_method(network,
+						CONNMAN_IPCONFIG_METHOD_DHCP);
+	} else {
 		struct connman_ipaddress *address;
 
 		if (config->ipv4_netmask == 0 ||
