@@ -56,6 +56,13 @@ DBusMessage *__connman_error_invalid_property(DBusMessage *msg);
 int __connman_manager_init(void);
 void __connman_manager_cleanup(void);
 
+enum time_updates {
+	TIME_UPDATES_UNKNOWN = 0,
+	TIME_UPDATES_MANUAL  = 1,
+	TIME_UPDATES_AUTO    = 2,
+};
+
+enum time_updates __connman_clock_timeupdates(void);
 int __connman_clock_init(void);
 void __connman_clock_cleanup(void);
 
@@ -133,6 +140,7 @@ int __connman_inet_modify_address(int cmd, int flags, int index, int family,
 				const char *peer,
 				unsigned char prefixlen,
 				const char *broadcast);
+int __connman_inet_get_interface_address(int index, int family, void *address);
 
 #include <netinet/ip6.h>
 #include <netinet/icmp6.h>
@@ -206,6 +214,7 @@ int __connman_storage_save_global(GKeyFile *keyfile);
 void __connman_storage_delete_global(void);
 
 GKeyFile *__connman_storage_load_config(const char *ident);
+GKeyFile *__connman_storage_load_provider_config(const char *ident);
 
 GKeyFile *__connman_storage_open_service(const char *ident);
 int __connman_storage_save_service(GKeyFile *keyfile, const char *ident);
@@ -361,6 +370,8 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 		GKeyFile *keyfile, const char *identifier, const char *prefix);
 gboolean __connman_ipconfig_ipv6_privacy_enabled(struct connman_ipconfig *ipconfig);
+int __connman_ipconfig_ipv6_set_privacy(struct connman_ipconfig *ipconfig,
+					const char *value);
 
 int __connman_ipconfig_set_rp_filter();
 void __connman_ipconfig_unset_rp_filter(int old_value);
@@ -556,7 +567,7 @@ int __connman_provider_indicate_state(struct connman_provider *provider,
 int __connman_provider_indicate_error(struct connman_provider *provider,
 					enum connman_provider_error error);
 int __connman_provider_connect(struct connman_provider *provider);
-int __connman_provider_remove(const char *path);
+int __connman_provider_remove_by_path(const char *path);
 void __connman_provider_cleanup(void);
 int __connman_provider_init(void);
 
@@ -600,8 +611,12 @@ int __connman_service_set_favorite_delayed(struct connman_service *service,
 					gboolean delay_ordering);
 int __connman_service_set_immutable(struct connman_service *service,
 						connman_bool_t immutable);
+int __connman_service_set_ignore(struct connman_service *service,
+						connman_bool_t ignore);
 void __connman_service_set_userconnect(struct connman_service *service,
 						connman_bool_t userconnect);
+void __connman_service_set_search_domains(struct connman_service *service,
+					char **domains);
 
 void __connman_service_set_string(struct connman_service *service,
 					const char *key, const char *value);
@@ -650,6 +665,8 @@ void __connman_service_nameserver_add_routes(struct connman_service *service,
 						const char *gw);
 void __connman_service_nameserver_del_routes(struct connman_service *service,
 					enum connman_ipconfig_type type);
+void __connman_service_set_timeservers(struct connman_service *service,
+						char **timeservers);
 int __connman_service_timeserver_append(struct connman_service *service,
 						const char *timeserver);
 int __connman_service_timeserver_remove(struct connman_service *service,
@@ -678,6 +695,9 @@ int __connman_service_set_passphrase(struct connman_service *service,
 const char *__connman_service_get_passphrase(struct connman_service *service);
 void __connman_service_set_agent_passphrase(struct connman_service *service,
 						const char *agent_passphrase);
+int __connman_service_reset_ipconfig(struct connman_service *service,
+		enum connman_ipconfig_type type, DBusMessageIter *array,
+		enum connman_service_state *new_state);
 
 void __connman_service_notify(struct connman_service *service,
 			unsigned int rx_packets, unsigned int tx_packets,
@@ -778,10 +798,34 @@ int __connman_stats_get(struct connman_service *service,
 				connman_bool_t roaming,
 				struct connman_stats_data *data);
 
+int __connman_iptables_dump(const char *table_name);
+int __connman_iptables_new_chain(const char *table_name,
+					const char *chain);
+int __connman_iptables_delete_chain(const char *table_name,
+					const char *chain);
+int __connman_iptables_flush_chain(const char *table_name,
+					const char *chain);
+int __connman_iptables_change_policy(const char *table_name,
+					const char *chain,
+					const char *policy);
+int __connman_iptables_append(const char *table_name,
+			const char *chain,
+			const char *rule_spec);
+int __connman_iptables_insert(const char *table_name,
+			const char *chain,
+			const char *rule_spec);
+int __connman_iptables_delete(const char *table_name,
+			const char *chain,
+			const char *rule_spec);
+
+typedef void (*connman_iptables_iterate_chains_cb_t) (const char *chain_name,
+							void *user_data);
+int __connman_iptables_iterate_chains(const char *table_name,
+				connman_iptables_iterate_chains_cb_t cb,
+				void *user_data);
+
 int __connman_iptables_init(void);
 void __connman_iptables_cleanup(void);
-int __connman_iptables_command(const char *format, ...)
-				__attribute__((format(printf, 1, 2)));
 int __connman_iptables_commit(const char *table_name);
 
 int __connman_dnsproxy_init(void);
@@ -843,3 +887,17 @@ void __connman_nat_cleanup(void);
 int __connman_nat_enable(const char *name, const char *address,
 				unsigned char prefixlen);
 void __connman_nat_disable(const char *name);
+
+struct firewall_context;
+
+struct firewall_context *__connman_firewall_create(void);
+void __connman_firewall_destroy(struct firewall_context *ctx);
+int __connman_firewall_add_rule(struct firewall_context *ctx,
+				const char *table,
+				const char *chain,
+				const char *rule_fmt, ...);
+int __connman_firewall_enable(struct firewall_context *ctx);
+int __connman_firewall_disable(struct firewall_context *ctx);
+
+int __connman_firewall_init(void);
+void __connman_firewall_cleanup(void);
