@@ -647,6 +647,36 @@ int __connman_device_disconnect(struct connman_device *device)
 	return 0;
 }
 
+int connman_device_disconnect_service(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	device->reconnect = FALSE;
+
+	if (device->network) {
+		struct connman_service *service =
+			connman_service_lookup_from_network(device->network);
+
+		if (service != NULL)
+			__connman_service_disconnect(service);
+		else
+			connman_network_set_connected(device->network, FALSE);
+	}
+
+	return 0;
+}
+
+int connman_device_reconnect_service(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	device->reconnect = TRUE;
+
+	__connman_service_auto_connect();
+
+	return 0;
+}
+
 static void mark_network_available(gpointer key, gpointer value,
 							gpointer user_data)
 {
@@ -1035,6 +1065,19 @@ struct connman_device *__connman_device_find_device(
 	return NULL;
 }
 
+struct connman_device *connman_device_find_by_index(int index)
+{
+	GSList *list;
+
+	for (list = device_list; list != NULL; list = list->next) {
+		struct connman_device *device = list->data;
+		if (device->index == index)
+			return device;
+	}
+
+	return NULL;
+}
+
 /**
  * connman_device_set_regdom
  * @device: device structure
@@ -1306,17 +1349,23 @@ connman_bool_t __connman_device_isfiltered(const char *devname)
 {
 	char **pattern;
 	char **blacklisted_interfaces;
+	gboolean match;
 
 	if (device_filter == NULL)
 		goto nodevice;
 
-	for (pattern = device_filter; *pattern; pattern++) {
-		if (g_pattern_match_simple(*pattern, devname) == FALSE) {
-			DBG("ignoring device %s (match)", devname);
-			return TRUE;
+	for (pattern = device_filter, match = FALSE; *pattern; pattern++) {
+		if (g_pattern_match_simple(*pattern, devname) == TRUE) {
+			match = TRUE;
+			break;
 		}
 	}
 
+	if (match == FALSE) {
+		DBG("ignoring device %s (match)", devname);
+		return TRUE;
+	}
+
 nodevice:
 	if (g_pattern_match_simple("dummy*", devname) == TRUE) {
 		DBG("ignoring dummy networking devices");
