@@ -50,6 +50,10 @@
 #define REN_MAX_RT      (600 * 1000)
 #define REB_TIMEOUT     (10 * 1000)
 #define REB_MAX_RT      (600 * 1000)
+#define CNF_MAX_DELAY   (1 * 1000)
+#define CNF_TIMEOUT	(1 * 1000)
+#define CNF_MAX_RT	(4 * 1000)
+#define CNF_MAX_RD	(10 * 1000)
 
 
 struct connman_dhcpv6 {
@@ -62,6 +66,7 @@ struct connman_dhcpv6 {
 	GDHCPClient *dhcp_client;
 
 	guint timeout;		/* operation timeout in msec */
+	guint MRD;		/* max operation timeout in msec */
 	guint RT;		/* in msec */
 	gboolean use_ta;	/* set to TRUE if IPv6 privacy is enabled */
 	GSList *prefixes;	/* network prefixes from radvd */
@@ -80,6 +85,11 @@ static void clear_timer(struct connman_dhcpv6 *dhcp)
 		g_source_remove(dhcp->timeout);
 		dhcp->timeout = 0;
 	}
+
+	if (dhcp->MRD > 0) {
+		g_source_remove(dhcp->MRD);
+		dhcp->MRD = 0;
+	}
 }
 
 static inline float get_random()
@@ -238,6 +248,10 @@ static void clear_callbacks(GDHCPClient *dhcp_client)
 				G_DHCP_CLIENT_EVENT_REQUEST,
 				NULL, NULL);
 
+	g_dhcp_client_register_event(dhcp_client,
+				G_DHCP_CLIENT_EVENT_CONFIRM,
+				NULL, NULL);
+
 	g_dhcp_client_register_event(dhcp_client,
 				G_DHCP_CLIENT_EVENT_RENEW,
 				NULL, NULL);
@@ -265,12 +279,14 @@ static void info_req_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 	DBG("dhcpv6 information-request %p", dhcp);
 
-	service = __connman_service_lookup_from_network(dhcp->network);
+	service = connman_service_lookup_from_network(dhcp->network);
 	if (service == NULL) {
 		connman_error("Can not lookup service");
 		return;
 	}
 
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
+
 	option = g_dhcp_client_get_option(dhcp_client, G_DHCPV6_DNS_SERVERS);
 	entries = g_list_length(option);
 
@@ -291,7 +307,8 @@ static void info_req_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 		dhcp->nameservers = nameservers;
 
-		for (i = 0; dhcp->nameservers[i] != NULL; i++)
+		for (i = 0; dhcp->nameservers != NULL &&
+					dhcp->nameservers[i] != NULL; i++)
 			__connman_service_nameserver_append(service,
 						dhcp->nameservers[i],
 						FALSE);
@@ -318,7 +335,8 @@ static void info_req_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 		dhcp->timeservers = timeservers;
 
-		for (i = 0; dhcp->timeservers[i] != NULL; i++)
+		for (i = 0; dhcp->timeservers != NULL &&
+					dhcp->timeservers[i] != NULL; i++)
 			__connman_service_timeserver_append(service,
 							dhcp->timeservers[i]);
 	} else
@@ -351,7 +369,7 @@ static int dhcpv6_info_request(struct connman_dhcpv6 *dhcp)
 	if (getenv("CONNMAN_DHCPV6_DEBUG"))
 		g_dhcp_client_set_debug(dhcp_client, dhcpv6_debug, "DHCPv6");
 
-	service = __connman_service_lookup_from_network(dhcp->network);
+	service = connman_service_lookup_from_network(dhcp->network);
 	if (service == NULL) {
 		clear_timer(dhcp);
 		g_dhcp_client_unref(dhcp_client);
@@ -437,7 +455,7 @@ static int set_addresses(GDHCPClient *dhcp_client,
 	const char *c_address;
 	char *address = NULL;
 
-	service = __connman_service_lookup_from_network(dhcp->network);
+	service = connman_service_lookup_from_network(dhcp->network);
 	if (service == NULL) {
 		connman_error("Can not lookup service");
 		return -EINVAL;
@@ -469,7 +487,8 @@ static int set_addresses(GDHCPClient *dhcp_client,
 
 		dhcp->nameservers = nameservers;
 
-		for (i = 0; dhcp->nameservers[i] != NULL; i++)
+		for (i = 0; dhcp->nameservers != NULL &&
+					dhcp->nameservers[i] != NULL; i++)
 			__connman_service_nameserver_append(service,
 							dhcp->nameservers[i],
 							FALSE);
@@ -496,7 +515,8 @@ static int set_addresses(GDHCPClient *dhcp_client,
 
 		dhcp->timeservers = timeservers;
 
-		for (i = 0; dhcp->timeservers[i] != NULL; i++)
+		for (i = 0; dhcp->timeservers != NULL &&
+					dhcp->timeservers[i] != NULL; i++)
 			__connman_service_timeserver_append(service,
 							dhcp->timeservers[i]);
 	} else
@@ -527,8 +547,13 @@ static int set_addresses(GDHCPClient *dhcp_client,
 		__connman_ipconfig_set_prefixlen(ipconfig, prefix_len);
 
 		DBG("new address %s/%d", address, prefix_len);
+
+		__connman_ipconfig_set_dhcp_address(ipconfig, address);
+		__connman_service_save(service);
 	}
 
+	g_free(address);
+
 	return 0;
 }
 
@@ -566,6 +591,7 @@ static void rebind_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 	g_dhcpv6_client_reset_rebind(dhcp_client);
 	g_dhcpv6_client_reset_renew(dhcp_client);
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
 
 	re_cb(dhcp_client, user_data);
 }
@@ -590,7 +616,7 @@ static int dhcpv6_rebind(struct connman_dhcpv6 *dhcp)
 	g_dhcpv6_client_set_ia(dhcp_client,
 			connman_network_get_index(dhcp->network),
 			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
-			NULL, NULL, FALSE);
+			NULL, NULL, FALSE, NULL);
 
 	clear_callbacks(dhcp_client);
 
@@ -648,6 +674,8 @@ static gboolean timeout_rebind(gpointer user_data)
 
 	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_rebind, dhcp);
 
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
 	g_dhcp_client_start(dhcp->dhcp_client, NULL);
 
 	return FALSE;
@@ -672,6 +700,8 @@ static void request_cb(GDHCPClient *dhcp_client, gpointer user_data)
 {
 	DBG("");
 
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
+
 	re_cb(dhcp_client, user_data);
 }
 
@@ -699,7 +729,7 @@ static int dhcpv6_request(struct connman_dhcpv6 *dhcp,
 	g_dhcpv6_client_set_ia(dhcp_client,
 			connman_network_get_index(dhcp->network),
 			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
-			&T1, &T2, add_addresses);
+			&T1, &T2, add_addresses, NULL);
 
 	clear_callbacks(dhcp_client);
 
@@ -729,6 +759,8 @@ static gboolean timeout_request(gpointer user_data)
 	DBG("request RT timeout %d msec", dhcp->RT);
 	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_request, dhcp);
 
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
 	g_dhcp_client_start(dhcp->dhcp_client, NULL);
 
 	return FALSE;
@@ -739,6 +771,7 @@ static void renew_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	DBG("");
 
 	g_dhcpv6_client_reset_renew(dhcp_client);
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
 
 	re_cb(dhcp_client, user_data);
 }
@@ -766,7 +799,7 @@ static int dhcpv6_renew(struct connman_dhcpv6 *dhcp)
 	g_dhcpv6_client_set_ia(dhcp_client,
 			connman_network_get_index(dhcp->network),
 			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
-			&T1, &T2, TRUE);
+			&T1, &T2, TRUE, NULL);
 
 	clear_callbacks(dhcp_client);
 
@@ -791,6 +824,8 @@ static gboolean timeout_renew(gpointer user_data)
 
 	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_renew, dhcp);
 
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
 	g_dhcp_client_start(dhcp->dhcp_client, NULL);
 
 	return FALSE;
@@ -918,7 +953,7 @@ int __connman_dhcpv6_start_release(struct connman_network *network,
 	g_dhcpv6_client_set_ia(dhcp_client,
 			connman_network_get_index(dhcp->network),
 			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
-			NULL, NULL, TRUE);
+			NULL, NULL, TRUE, NULL);
 
 	clear_callbacks(dhcp_client);
 
@@ -975,6 +1010,8 @@ static gboolean timeout_info_req(gpointer user_data)
 
 	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_info_req, dhcp);
 
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
 	g_dhcp_client_start(dhcp->dhcp_client, NULL);
 
 	return FALSE;
@@ -1041,6 +1078,8 @@ static void advertise_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 	clear_timer(dhcp);
 
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
+
 	if (g_dhcpv6_client_get_status(dhcp_client) != 0) {
 		if (dhcp->callback != NULL)
 			dhcp->callback(dhcp->network, FALSE);
@@ -1066,6 +1105,8 @@ static void solicitation_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	clear_timer(dhcp);
 
 	set_addresses(dhcp_client, dhcp);
+
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
 }
 
 static gboolean timeout_solicitation(gpointer user_data)
@@ -1078,6 +1119,8 @@ static gboolean timeout_solicitation(gpointer user_data)
 
 	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_solicitation, dhcp);
 
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
 	g_dhcp_client_start(dhcp->dhcp_client, NULL);
 
 	return FALSE;
@@ -1104,7 +1147,7 @@ static int dhcpv6_solicitation(struct connman_dhcpv6 *dhcp)
 	if (getenv("CONNMAN_DHCPV6_DEBUG"))
 		g_dhcp_client_set_debug(dhcp_client, dhcpv6_debug, "DHCPv6");
 
-	service = __connman_service_lookup_from_network(dhcp->network);
+	service = connman_service_lookup_from_network(dhcp->network);
 	if (service == NULL) {
 		clear_timer(dhcp);
 		g_dhcp_client_unref(dhcp_client);
@@ -1131,7 +1174,7 @@ static int dhcpv6_solicitation(struct connman_dhcpv6 *dhcp)
 
 	g_dhcpv6_client_set_ia(dhcp_client, index,
 			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
-			NULL, NULL, FALSE);
+			NULL, NULL, FALSE, NULL);
 
 	clear_callbacks(dhcp_client);
 
@@ -1164,10 +1207,150 @@ static gboolean start_solicitation(gpointer user_data)
 	return FALSE;
 }
 
+static void confirm_cb(GDHCPClient *dhcp_client, gpointer user_data)
+{
+	struct connman_dhcpv6 *dhcp = user_data;
+	int status = g_dhcpv6_client_get_status(dhcp_client);
+
+	DBG("dhcpv6 confirm msg %p status %d", dhcp, status);
+
+	clear_timer(dhcp);
+
+	set_addresses(dhcp_client, dhcp);
+
+	g_dhcpv6_client_clear_retransmit(dhcp_client);
+
+	/*
+	 * If confirm fails, start from scratch.
+	 */
+	if (status != 0) {
+		g_dhcp_client_unref(dhcp->dhcp_client);
+		start_solicitation(dhcp);
+	} else if (dhcp->callback != NULL)
+		dhcp->callback(dhcp->network, TRUE);
+}
+
+static int dhcpv6_confirm(struct connman_dhcpv6 *dhcp)
+{
+	GDHCPClient *dhcp_client;
+	GDHCPClientError error;
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig_ipv6;
+	int index, ret;
+
+	DBG("dhcp %p", dhcp);
+
+	index = connman_network_get_index(dhcp->network);
+
+	dhcp_client = g_dhcp_client_new(G_DHCP_IPV6, index, &error);
+	if (error != G_DHCP_CLIENT_ERROR_NONE) {
+		clear_timer(dhcp);
+		return -EINVAL;
+	}
+
+	if (getenv("CONNMAN_DHCPV6_DEBUG"))
+		g_dhcp_client_set_debug(dhcp_client, dhcpv6_debug, "DHCPv6");
+
+	service = connman_service_lookup_from_network(dhcp->network);
+	if (service == NULL) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
+		return -EINVAL;
+	}
+
+	ret = set_duid(service, dhcp->network, dhcp_client, index);
+	if (ret < 0) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
+		return ret;
+	}
+
+	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_CLIENTID);
+	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_RAPID_COMMIT);
+	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_DNS_SERVERS);
+	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_SNTP_SERVERS);
+
+	g_dhcpv6_client_set_oro(dhcp_client, 2, G_DHCPV6_DNS_SERVERS,
+				G_DHCPV6_SNTP_SERVERS);
+
+	ipconfig_ipv6 = __connman_service_get_ip6config(service);
+	dhcp->use_ta = __connman_ipconfig_ipv6_privacy_enabled(ipconfig_ipv6);
+
+	g_dhcpv6_client_set_ia(dhcp_client, index,
+			dhcp->use_ta == TRUE ? G_DHCPV6_IA_TA : G_DHCPV6_IA_NA,
+			NULL, NULL, TRUE,
+			__connman_ipconfig_get_dhcp_address(ipconfig_ipv6));
+
+	clear_callbacks(dhcp_client);
+
+	g_dhcp_client_register_event(dhcp_client,
+				G_DHCP_CLIENT_EVENT_CONFIRM,
+				confirm_cb, dhcp);
+
+	dhcp->dhcp_client = dhcp_client;
+
+	return g_dhcp_client_start(dhcp_client, NULL);
+}
+
+static gboolean timeout_confirm(gpointer user_data)
+{
+	struct connman_dhcpv6 *dhcp = user_data;
+
+	dhcp->RT = calc_delay(dhcp->RT, CNF_MAX_RT);
+
+	DBG("confirm RT timeout %d msec", dhcp->RT);
+
+	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_confirm, dhcp);
+
+	g_dhcpv6_client_set_retransmit(dhcp->dhcp_client);
+
+	g_dhcp_client_start(dhcp->dhcp_client, NULL);
+
+	return FALSE;
+}
+
+static gboolean timeout_max_confirm(gpointer user_data)
+{
+	struct connman_dhcpv6 *dhcp = user_data;
+
+	dhcp->MRD = 0;
+
+	clear_timer(dhcp);
+
+	DBG("confirm max retransmit duration timeout");
+
+	g_dhcpv6_client_clear_retransmit(dhcp->dhcp_client);
+
+	if (dhcp->callback != NULL)
+		dhcp->callback(dhcp->network, FALSE);
+
+	return FALSE;
+}
+
+static gboolean start_confirm(gpointer user_data)
+{
+	struct connman_dhcpv6 *dhcp = user_data;
+
+	/* Set the confirm timeout, RFC 3315 chapter 14 */
+	dhcp->RT = CNF_TIMEOUT * (1 + get_random());
+
+	DBG("confirm initial RT timeout %d msec", dhcp->RT);
+
+	dhcp->timeout = g_timeout_add(dhcp->RT, timeout_confirm, dhcp);
+	dhcp->MRD = g_timeout_add(CNF_MAX_RD, timeout_max_confirm, dhcp);
+
+	dhcpv6_confirm(dhcp);
+
+	return FALSE;
+}
+
 int __connman_dhcpv6_start(struct connman_network *network,
 				GSList *prefixes, dhcp_cb callback)
 {
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig_ipv6;
 	struct connman_dhcpv6 *dhcp;
+	char *last_address;
 	int delay;
 
 	DBG("");
@@ -1178,6 +1361,10 @@ int __connman_dhcpv6_start(struct connman_network *network,
 			return -EBUSY;
 	}
 
+	service = connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return -EINVAL;
+
 	dhcp = g_try_new0(struct connman_dhcpv6, 1);
 	if (dhcp == NULL)
 		return -ENOMEM;
@@ -1196,7 +1383,24 @@ int __connman_dhcpv6_start(struct connman_network *network,
 	/* Initial timeout, RFC 3315, 17.1.2 */
 	delay = rand() % 1000;
 
-	dhcp->timeout = g_timeout_add(delay, start_solicitation, dhcp);
+	ipconfig_ipv6 = __connman_service_get_ip6config(service);
+	last_address = __connman_ipconfig_get_dhcp_address(ipconfig_ipv6);
+
+	if (prefixes != NULL && last_address != NULL &&
+			check_ipv6_addr_prefix(prefixes,
+						last_address) != 128) {
+		/*
+		 * So we are in the same subnet
+		 * RFC 3315, chapter 18.1.2 Confirm message
+		 */
+		dhcp->timeout = g_timeout_add(delay, start_confirm, dhcp);
+	} else {
+		/*
+		 * Start from scratch.
+		 * RFC 3315, chapter 17.1.2 Solicitation message
+		 */
+		dhcp->timeout = g_timeout_add(delay, start_solicitation, dhcp);
+	}
 
 	return 0;
 }
