@@ -391,18 +391,6 @@ static void __connman_ipconfig_lower_up(struct connman_ipdevice *ipdevice)
 {
 	DBG("ipconfig ipv4 %p ipv6 %p", ipdevice->config_ipv4,
 					ipdevice->config_ipv6);
-
-	if (ipdevice->config_ipv6 != NULL &&
-			ipdevice->config_ipv6->enabled == TRUE)
-		return;
-
-	if (ipdevice->ifname == NULL)
-		return;
-
-	if (__connman_device_isfiltered(ipdevice->ifname) == FALSE) {
-		ipdevice->ipv6_enabled = get_ipv6_state(ipdevice->ifname);
-		set_ipv6_state(ipdevice->ifname, FALSE);
-	}
 }
 
 static void __connman_ipconfig_lower_down(struct connman_ipdevice *ipdevice)
@@ -675,11 +663,11 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 		__connman_ippool_newaddr(index, address, prefixlen);
 
 	if (ipdevice->config_ipv4 != NULL && family == AF_INET)
-		connman_ipaddress_copy(ipdevice->config_ipv4->system,
+		connman_ipaddress_copy_address(ipdevice->config_ipv4->system,
 					ipaddress);
 
 	else if (ipdevice->config_ipv6 != NULL && family == AF_INET6)
-		connman_ipaddress_copy(ipdevice->config_ipv6->system,
+		connman_ipaddress_copy_address(ipdevice->config_ipv6->system,
 					ipaddress);
 	else
 		return;
@@ -1568,7 +1556,6 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 		if (ipdevice->config_ipv6 == ipconfig)
 			return -EALREADY;
 		type = CONNMAN_IPCONFIG_TYPE_IPV6;
-		enable_ipv6(ipconfig);
 	} else
 		return -EINVAL;
 
@@ -1596,9 +1583,11 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
 		ipdevice->config_ipv4 = __connman_ipconfig_ref(ipconfig);
-	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6) {
 		ipdevice->config_ipv6 = __connman_ipconfig_ref(ipconfig);
 
+		enable_ipv6(ipdevice->config_ipv6);
+	}
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
 	if (ipdevice->flags & IFF_UP)
@@ -1729,6 +1718,25 @@ static int string2privacy(const char *privacy)
 		return 0;
 }
 
+int __connman_ipconfig_ipv6_set_privacy(struct connman_ipconfig *ipconfig,
+					const char *value)
+{
+	int privacy;
+
+	if (ipconfig == NULL)
+		return -EINVAL;
+
+	DBG("ipconfig %p privacy %s", ipconfig, value);
+
+	privacy = string2privacy(value);
+
+	ipconfig->ipv6_privacy_config = privacy;
+
+	enable_ipv6(ipconfig);
+
+	return 0;
+}
+
 void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 							DBusMessageIter *iter)
 {
@@ -1944,6 +1952,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		*privacy_string = NULL;
 	int prefix_length = 0, privacy = 0;
 	DBusMessageIter dict;
+	int type = -1;
 
 	DBG("ipconfig %p", ipconfig);
 
@@ -2032,7 +2041,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		if (ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
-			return -EINVAL;
+			return -EOPNOTSUPP;
 
 		ipconfig->method = method;
 		if (privacy_string != NULL)
@@ -2041,7 +2050,26 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		break;
 
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
-		if (address == NULL)
+		switch (ipconfig->type) {
+		case CONNMAN_IPCONFIG_TYPE_IPV4:
+			type = AF_INET;
+			break;
+		case CONNMAN_IPCONFIG_TYPE_IPV6:
+			type = AF_INET6;
+			break;
+		case CONNMAN_IPCONFIG_TYPE_UNKNOWN:
+			type = -1;
+			break;
+		}
+
+		if ((address != NULL && connman_inet_check_ipaddress(address)
+						!= type) ||
+				(netmask != NULL &&
+				connman_inet_check_ipaddress(netmask)
+						!= type) ||
+				(gateway != NULL &&
+				connman_inet_check_ipaddress(gateway)
+						!= type))
 			return -EINVAL;
 
 		ipconfig->method = method;
@@ -2056,7 +2084,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		break;
 
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
-		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		if (ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV4)
 			return -EOPNOTSUPP;
 
 		ipconfig->method = method;
@@ -2194,6 +2222,15 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 		key = g_strdup_printf("%sprivacy", prefix);
 		g_key_file_set_string(keyfile, identifier, key, privacy);
 		g_free(key);
+
+		key = g_strdup_printf("%sDHCP.LastAddress", prefix);
+		if (ipconfig->last_dhcp_address != NULL &&
+				strlen(ipconfig->last_dhcp_address) > 0)
+			g_key_file_set_string(keyfile, identifier, key,
+					ipconfig->last_dhcp_address);
+		else
+			g_key_file_remove_key(keyfile, identifier, key, NULL);
+		g_free(key);
 	}
 
 	switch (ipconfig->method) {
