@@ -50,8 +50,8 @@ struct connman_network {
 	connman_bool_t available;
 	connman_bool_t connected;
 	connman_bool_t roaming;
-	connman_uint8_t strength;
-	connman_uint16_t frequency;
+	uint8_t strength;
+	uint16_t frequency;
 	char *identifier;
 	char *name;
 	char *node;
@@ -1769,7 +1769,7 @@ int connman_network_set_name(struct connman_network *network,
  */
 
 int connman_network_set_strength(struct connman_network *network,
-						connman_uint8_t strength)
+						uint8_t strength)
 {
 	DBG("network %p strengh %d", network, strength);
 
@@ -1778,13 +1778,13 @@ int connman_network_set_strength(struct connman_network *network,
 	return 0;
 }
 
-connman_uint8_t connman_network_get_strength(struct connman_network *network)
+uint8_t connman_network_get_strength(struct connman_network *network)
 {
 	return network->strength;
 }
 
 int connman_network_set_frequency(struct connman_network *network,
-						connman_uint16_t frequency)
+						uint16_t frequency)
 {
 	DBG("network %p frequency %d", network, frequency);
 
@@ -1793,13 +1793,13 @@ int connman_network_set_frequency(struct connman_network *network,
 	return 0;
 }
 
-connman_uint16_t connman_network_get_frequency(struct connman_network *network)
+uint16_t connman_network_get_frequency(struct connman_network *network)
 {
 	return network->frequency;
 }
 
 int connman_network_set_wifi_channel(struct connman_network *network,
-						connman_uint16_t channel)
+						uint16_t channel)
 {
 	DBG("network %p wifi channel %d", network, channel);
 
@@ -1808,7 +1808,7 @@ int connman_network_set_wifi_channel(struct connman_network *network,
 	return 0;
 }
 
-connman_uint16_t connman_network_get_wifi_channel(struct connman_network *network)
+uint16_t connman_network_get_wifi_channel(struct connman_network *network)
 {
 	return network->wifi.channel;
 }
