@@ -35,6 +35,8 @@
 #include <string.h>
 #include <fcntl.h>
 #include <linux/if_tun.h>
+#include <netinet/in.h>
+#include <linux/if_bridge.h>
 
 #include "connman.h"
 
@@ -46,15 +48,12 @@
 #define DBUS_TYPE_UNIX_FD -1
 #endif
 
-#define BRIDGE_PROC_DIR "/proc/sys/net/bridge"
-
 #define BRIDGE_NAME "tether"
-#define BRIDGE_DNS "8.8.8.8"
 
 #define DEFAULT_MTU	1500
 
-#define PRIVATE_NETWORK_PRIMARY_DNS BRIDGE_DNS
-#define PRIVATE_NETWORK_SECONDARY_DNS "8.8.4.4"
+static char *private_network_primary_dns = NULL;
+static char *private_network_secondary_dns = NULL;
 
 static volatile int tethering_enabled;
 static GDHCPServer *tethering_dhcp_server = NULL;
@@ -73,15 +72,24 @@ struct connman_private_network {
 	int index;
 	guint iface_watch;
 	struct connman_ippool *pool;
-	const char *primary_dns;
-	const char *secondary_dns;
+	char *primary_dns;
+	char *secondary_dns;
 };
 
 const char *__connman_tethering_get_bridge(void)
 {
-	struct stat st;
+	int sk, err;
+	unsigned long args[3];
 
-	if (stat(BRIDGE_PROC_DIR, &st) < 0) {
+	sk = socket(AF_INET, SOCK_STREAM, 0);
+	if (sk < 0)
+		return NULL;
+
+	args[0] = BRCTL_GET_VERSION;
+	args[1] = args[2] = 0;
+	err = ioctl(sk, SIOCGIFBR, &args);
+	close(sk);
+	if (err == -1) {
 		connman_error("Missing support for 802.1d ethernet bridging");
 		return NULL;
 	}
@@ -183,6 +191,7 @@ void __connman_tethering_set_enabled(void)
 	const char *end_ip;
 	const char *dns;
 	unsigned char prefixlen;
+	char **ns;
 
 	DBG("enabled %d", tethering_enabled + 1);
 
@@ -190,14 +199,18 @@ void __connman_tethering_set_enabled(void)
 		return;
 
 	err = __connman_bridge_create(BRIDGE_NAME);
-	if (err < 0)
+	if (err < 0) {
+		__sync_fetch_and_sub(&tethering_enabled, 1);
 		return;
+	}
 
 	index = connman_inet_ifindex(BRIDGE_NAME);
 	dhcp_ippool = __connman_ippool_create(index, 2, 252,
 						tethering_restart, NULL);
 	if (dhcp_ippool == NULL) {
 		connman_error("Fail to create IP pool");
+		__connman_bridge_remove(BRIDGE_NAME);
+		__sync_fetch_and_sub(&tethering_enabled, 1);
 		return;
 	}
 
@@ -209,15 +222,34 @@ void __connman_tethering_set_enabled(void)
 
 	err = __connman_bridge_enable(BRIDGE_NAME, gateway, broadcast);
 	if (err < 0 && err != -EALREADY) {
+		__connman_ippool_unref(dhcp_ippool);
 		__connman_bridge_remove(BRIDGE_NAME);
+		__sync_fetch_and_sub(&tethering_enabled, 1);
 		return;
 	}
 
+	ns = connman_setting_get_string_list("FallbackNameservers");
+	if (ns != NULL) {
+		if (ns[0] != NULL) {
+			g_free(private_network_primary_dns);
+			private_network_primary_dns = g_strdup(ns[0]);
+		}
+		if (ns[1] != NULL) {
+			g_free(private_network_secondary_dns);
+			private_network_secondary_dns = g_strdup(ns[1]);
+		}
+
+		DBG("Fallback ns primary %s secondary %s",
+			private_network_primary_dns,
+			private_network_secondary_dns);
+	}
+
 	dns = gateway;
-	if (__connman_dnsproxy_add_listener(BRIDGE_NAME) < 0) {
+	if (__connman_dnsproxy_add_listener(index) < 0) {
 		connman_error("Can't add listener %s to DNS proxy",
 								BRIDGE_NAME);
-		dns = BRIDGE_DNS;
+		dns = private_network_primary_dns;
+		DBG("Serving %s nameserver to clients", dns);
 	}
 
 	tethering_dhcp_server = dhcp_server_start(BRIDGE_NAME,
@@ -226,12 +258,14 @@ void __connman_tethering_set_enabled(void)
 						24 * 3600, dns);
 	if (tethering_dhcp_server == NULL) {
 		__connman_bridge_disable(BRIDGE_NAME);
+		__connman_ippool_unref(dhcp_ippool);
 		__connman_bridge_remove(BRIDGE_NAME);
+		__sync_fetch_and_sub(&tethering_enabled, 1);
 		return;
 	}
 
 	prefixlen =
-		__connman_ipconfig_netmask_prefix_len(subnet_mask);
+		__connman_ipaddress_netmask_prefix_len(subnet_mask);
 	__connman_nat_enable(BRIDGE_NAME, start_ip, prefixlen);
 
 	DBG("tethering started");
@@ -239,9 +273,12 @@ void __connman_tethering_set_enabled(void)
 
 void __connman_tethering_set_disabled(void)
 {
+	int index;
+
 	DBG("enabled %d", tethering_enabled - 1);
 
-	__connman_dnsproxy_remove_listener(BRIDGE_NAME);
+	index = connman_inet_ifindex(BRIDGE_NAME);
+	__connman_dnsproxy_remove_listener(index);
 
 	if (__sync_fetch_and_sub(&tethering_enabled, 1) != 1)
 		return;
@@ -258,6 +295,11 @@ void __connman_tethering_set_disabled(void)
 
 	__connman_bridge_remove(BRIDGE_NAME);
 
+	g_free(private_network_primary_dns);
+	private_network_primary_dns = NULL;
+	g_free(private_network_secondary_dns);
+	private_network_secondary_dns = NULL;
+
 	DBG("tethering stopped");
 }
 
@@ -281,7 +323,7 @@ static void setup_tun_interface(unsigned int flags, unsigned change,
 	server_ip = __connman_ippool_get_start_ip(pn->pool);
 	peer_ip = __connman_ippool_get_end_ip(pn->pool);
 	prefixlen =
-		__connman_ipconfig_netmask_prefix_len(subnet_mask);
+		__connman_ipaddress_netmask_prefix_len(subnet_mask);
 
 	if ((__connman_inet_modify_address(RTM_NEWADDR,
 				NLM_F_REPLACE | NLM_F_ACK, pn->index, AF_INET,
@@ -309,9 +351,12 @@ static void setup_tun_interface(unsigned int flags, unsigned change,
 					DBUS_TYPE_STRING, &server_ip);
 	connman_dbus_dict_append_basic(&dict, "PeerIPv4",
 					DBUS_TYPE_STRING, &peer_ip);
-	connman_dbus_dict_append_basic(&dict, "PrimaryDNS",
+	if (pn->primary_dns != NULL)
+		connman_dbus_dict_append_basic(&dict, "PrimaryDNS",
 					DBUS_TYPE_STRING, &pn->primary_dns);
-	connman_dbus_dict_append_basic(&dict, "SecondaryDNS",
+
+	if (pn->secondary_dns != NULL)
+		connman_dbus_dict_append_basic(&dict, "SecondaryDNS",
 					DBUS_TYPE_STRING, &pn->secondary_dns);
 
 	connman_dbus_dict_close(&array, &dict);
@@ -347,6 +392,8 @@ static void remove_private_network(gpointer user_data)
 	g_free(pn->interface);
 	g_free(pn->owner);
 	g_free(pn->path);
+	g_free(pn->primary_dns);
+	g_free(pn->secondary_dns);
 	g_free(pn);
 }
 
@@ -427,8 +474,8 @@ int __connman_private_network_request(DBusMessage *msg, const char *owner)
 		goto error;
 	}
 
-	pn->primary_dns = PRIVATE_NETWORK_PRIMARY_DNS;
-	pn->secondary_dns = PRIVATE_NETWORK_SECONDARY_DNS;
+	pn->primary_dns = g_strdup(private_network_primary_dns);
+	pn->secondary_dns = g_strdup(private_network_secondary_dns);
 
 	pn->iface_watch = connman_rtnl_add_newlink_watch(index,
 						setup_tun_interface, pn);
