@@ -65,7 +65,6 @@ static void save_timeservers(char **servers)
 static char **load_timeservers()
 {
 	GKeyFile *keyfile;
-	GError *error = NULL;
 	char **servers = NULL;
 
 	keyfile = __connman_storage_load_global();
@@ -73,11 +72,7 @@ static char **load_timeservers()
 		return NULL;
 
 	servers = g_key_file_get_string_list(keyfile, "global",
-						"Timeservers", NULL, &error);
-	if (error) {
-		DBG("Error loading timeservers: %s", error->message);
-		g_error_free(error);
-	}
+						"Timeservers", NULL, NULL);
 
 	g_key_file_free(keyfile);
 
@@ -92,8 +87,16 @@ static void resolv_result(GResolvResultStatus status, char **results, gpointer u
 
 	if (status == G_RESOLV_RESULT_STATUS_SUCCESS) {
 		if (results != NULL) {
-			for (i = 0; results[i]; i++)
-				DBG("result: %s", results[i]);
+			for (i = 0; results[i]; i++) {
+				DBG("result[%d]: %s", i, results[i]);
+				if (i == 0)
+					continue;
+
+				ts_list = __connman_timeserver_add_list(
+							ts_list, results[i]);
+			}
+
+			DBG("Using timeserver %s", results[0]);
 
 			__connman_ntp_start(results[0]);
 
@@ -115,9 +118,6 @@ static void resolv_result(GResolvResultStatus status, char **results, gpointer u
 void __connman_timeserver_sync_next()
 {
 	char *server;
-	int ret;
-	struct addrinfo hints;
-	struct addrinfo *addr;
 
 	__connman_ntp_stop();
 
@@ -129,16 +129,9 @@ void __connman_timeserver_sync_next()
 
 	ts_list = g_slist_delete_link(ts_list, ts_list);
 
-	memset(&hints, 0, sizeof(struct addrinfo));
-	hints.ai_flags = AI_NUMERICHOST;
-	addr = NULL;
-
-	ret = getaddrinfo(server, NULL, &hints, &addr);
-	freeaddrinfo(addr);
-
 	/* if its a IP , directly query it. */
-	if (ret == 0) {
-		DBG("Using timeservers %s", server);
+	if (connman_inet_check_ipaddress(server) > 0) {
+		DBG("Using timeserver %s", server);
 
 		__connman_ntp_start(server);
 
@@ -189,6 +182,9 @@ GSList *__connman_timeserver_get_all(struct connman_service *service)
 	char **fallback_ts;
 	int index, i;
 
+	if (__connman_clock_timeupdates() == TIME_UPDATES_MANUAL)
+		return NULL;
+
 	service_ts_config = connman_service_get_timeservers_config(service);
 
 	/* First add Service Timeservers.Configuration to the list */
